from typing import List
from collections import Counter
from collections.abc import Mapping, Sequence
import numpy as np

from numbers import Number
import math

import operator
import itertools

import pprint

from functools import reduce


def to_dicts(s):
  if isinstance(s, str):
    return s
  if isinstance(s, Mapping):
    return {k: to_dicts(v) for k, v in s.items()}
  if isinstance(s, Sequence):
    return s.__class__([to_dicts(v) for v in s])
  else:
    return s


def to_structs(d):
  if isinstance(d, str):
    return d
  if isinstance(d, dict):
    return Struct({k: to_structs(v) for k, v in d.items()})
  if isinstance(d, Sequence):
    return d.__class__([to_structs(v) for v in d])
  else:
    return d


class Indexed():
  def __init__(self, struct):
    assert isinstance(struct, Struct)
    self._struct = struct

  def __getitem__(self, index):
    return self._struct._map(lambda x: x[index])


class Struct(Mapping):
  def __init__(self, entries):
    assert isinstance(entries, dict)
    self._entries = entries

  @staticmethod
  def build(d):
    return Struct(d)

  @staticmethod
  def create(**d):
    return Struct(d)

  def __reduce__(self):
    return (self.__class__, (self._entries,))

  def __setitem__(self, index, value):
    self._entries[index] = value

  def __getitem__(self, index):
    return self._entries[index]

  @property
  def _index(self) -> np.ndarray:
    return Indexed(self)

  def __iter__(self):
    return self._entries.__iter__()

  def items(self):
    return self._entries.items()

  def keys(self):
    return self._entries.keys()

  def values(self):
    return self._entries.values()

  def __getattr__(self, k):
    if k[0] == "_":
      return object.__getattribute__(self, k)
    elif k in self._entries:
      return self._entries[k]
    else:
      raise AttributeError(
          f"Struct does not contain '{k}', options are {list(self._entries.keys())}")

  def __setattr__(self, k, v):
    if k[0] == "_":
      return object.__setattr__(self, k, v)
    else:
      self._entries[k] = v


  def __eq__(self, other):
    if isinstance(other, Struct):
      return self._entries == other._entries
    else:
      return False

  def _to_dicts(self):
    return to_dicts(self)

  def _subset(self, *keys):
    d = {k: self[k] for k in keys}
    return self.build(d)

  def _without(self, *keys):
    d = {k: v for k, v in self.items() if not (k in keys)}
    return self.build(d)

  def _filter_none(self):
    return self.build({k: v for k, v in self.items() if v is not None})

  def _filterWithKey(self, f, *args, **kwargs):
    return self.build({k: v for k, v in self.items() if f(k, *args, **kwargs)})


  def _filterMapWithKey(self, f, *args, **kwargs):
    return self.build({k: result for k, v in self.items()
                       for result in [f(k, v, *args, **kwargs)]
                       if result is not None
                       })

  def _filterMap(self, f, *args, **kwargs):
    return self.build({k: result for k, v in self.items()
                       for result in [f(v, *args, **kwargs)]
                       if result is not None
                       })

  def _map(self, f, *args, **kwargs):
    d = {k: f(v, *args, **kwargs) for k, v in self.items()}
    return self.build(d)

  def _mapWithKey(self, f):
    m = {k: f(k, v) for k, v in self._entries.items()}
    return self.build(m)

  def __repr__(self):
    commaSep = ", ".join(["{}={}".format(str(k), repr(v))
                          for k, v in self.items()])
    return "{" + commaSep + "}"

  def __str__(self):
    return self.__repr__()

  def __len__(self):
    return self._entries.__len__()

  def __floordiv__(self, other):
    if isinstance(other, Number):
      return self._map(operator.floordiv, other)
    else:
      return self._zipWith(operator.floordiv, other)

  def __truediv__(self, other):
    if isinstance(other, Number):
      return self._map(operator.truediv, other)
    else:
      return self._zipWith(operator.truediv, other)

  def __add__(self, other):
    if isinstance(other, Number):
      return self._map(operator.add, other)
    else:
      return self._zipWith(operator.add, other)

  def __mul__(self, other):
    if isinstance(other, Number):
      return self._map(operator.mul, other)
    else:
      return self._zipWith(operator.mul, other)

  def _zipWith(self, f, other):
    assert isinstance(other, Struct)

    r = {k: f(self[k], other[k]) for k in self.keys()}
    return self.build(r)

  def _merge(self, other):
    """
    returns a struct which is a merge of this struct and another.
    """

    assert isinstance(other, Struct)
    d = self._entries.copy()
    d.update(other._entries)

    return self.build(d)

  def _extend(self, **values):
    d = self._entries.copy()
    d.update(values)

    return self.build(d)

  def _update(self, **values):
    for k in values:
      assert k in self, f"_update: entry not found {k}, options are {list[self.keys()]}"

    return self._extend(**values)

  def __radd__(self, other):
    return self.__add__(other)

  def __rmul__(self, other):
    return self.__mul__(other)


def set_recursive(path, d, value):
  if len(path) == 0:
    return value

  head, *rest = path
  if isinstance(d, Struct):
    update = {head: set_recursive(rest, d[head], value)}
    up = d._update(**update)
    return up

  assert False


class Lens(object):
  def __init__(self, path):
    self.path = path

  def __getattr__(self, k):
    return Lens(self.path + [k])

  def set(self, d, value):
    return set_recursive(self.path, d, value)


lens = Lens([])


class ZipList():
  def __init__(self, elems):
    self.elems = list(elems)

  def __getitem__(self, index):
    return self.elems[index]

  def __iter__(self):
    return self.elems.__iter__()

  def __eq__(self, other):
    if isinstance(other, ZipList):
      return self.elems == other.elems
    else:
      return False

  def __repr__(self):
    return self.elems.__repr__()

  def __str__(self):
    return self.elems.__str__()

  def __len__(self):
    return self.elems.__len__()

  def __floordiv__(self, other):
    if isinstance(other, Number):
      return self._map(operator.floordiv, other)
    else:
      return self._zipWith(operator.floordiv, other)

  def __truediv__(self, other):
    if isinstance(other, Number):
      return self._map(operator.truediv, other)
    else:
      return self._zipWith(operator.truediv, other)

  def _map(self, f, *args, **kwargs):
    return ZipList([f(v, *args, **kwargs) for v in self.elems])

  def _zipWith(self, f, other):

    assert isinstance(other, ZipList)
    assert len(self) == len(other)

    r = [f(x, y) for x, y in zip(self.elems, other.elems)]
    return ZipList(r)

  def __add__(self, other):
    if isinstance(other, Number):
      return self._map(operator.add, other)
    else:
      return self._zipWith(operator.add, other)

  def __mul__(self, other):
    if isinstance(other, Number):
      return self._map(operator.mul, other)
    else:
      return self._zipWith(operator.mul, other)

  def __radd__(self, other):
    return self.__add__(other)

  def __rmul__(self, other):
    return self.__mul__(other)

  def append(x):
    self.elems.append(x)
    return self


def struct(**d):
  return Struct(d)


def flatten(x, prefix=''):
  def add_prefix(k):
    if prefix == '':
      return str(k)
    else:
      return prefix + "." + str(k)

  def flatten_iter(iter):
    return [flatten(inner, add_prefix(i)) for i, inner in iter]

  if type(x) == list:
    return flatten_iter(enumerate(x))
  elif type(x) == tuple:
    return flatten_iter(enumerate(x))
  elif isinstance(x, Mapping):
    return flatten_iter(x.items())
  else:
    return [(prefix, x)]


def get_default_args(func):
  """
  returns a dictionary of arg_name:default_values for the input function
  """
  args, varargs, keywords, defaults = inspect.getargspec(func)
  return dict(zip(reversed(args), reversed(defaults)))


def replace(d, key, value):
  return {**d, key: value}


def over_struct(key, f):
  def modify(d):
    value = f(d[key])
    return Struct(replace(d, key, value))
  return modify


def over(key, f):
  def modify(d):
    value = f(d[key])
    return replace(d, key, value)
  return modify


def transpose_partial(dicts):
  accum = {}
  for d in dicts:
    for k, v in d.items():
      if k in accum:
        accum[k].append(v)
      else:
        accum[k] = [v]
  return accum


def transpose_partial_structs(structs):
  return Struct(transpose_partial(d.__dict__ for d in structs))


def transpose_structs(structs):
  elem = structs[0]
  d = {key: [d[key] for d in structs] for key in elem.keys()}
  return Struct(d)


def transpose_lists(lists):
  return list(zip(*lists))


def split_list(xs, sizes):
  assert len(xs) == sum(sizes)
  splits = []
  for size in sizes:
    
    splits.append(xs[:size])
    xs = xs[size:] 
  return splits

def split_dict(d):
  return list(d.keys()), list(d.values())

def invert_keys(d):
  inverted_d = {v:k for k, v in d.items()}
  return d.__class__(inverted_d)

def subset(d, keys):
  d_subset = {k:d[k] for k in keys}
  return d.__class__(d_subset)

def drop_while(f, xs):
  while(len(xs) > 0 and f(xs[0])):
    _, *xs = xs

  return xs


def filter_none(xs):
  return [x for x in xs if x is not None]


def filter_map(f, xs):
  return filter_none(map(f, xs))


def pluck(k, xs, default=None):
  return [d.get(k, default) for d in xs]


def pluck_struct(k, xs, default=None):
  return xs._map(lambda x: x.get(k, default))


def const(x):
  def f(*y):
    return x
  return f


def concat_lists(xs):
  return list(itertools.chain.from_iterable(xs))


def map_dict(f, d, *args, **kwargs):
  return {k:  f(v,  *args, **kwargs) for k, v in d.items()}


def map_list(f, xs, *args, **kwargs):
  return [f(x, *args, **kwargs) for x in xs]


def map_none(f, x, *args, **kwargs):
  return f(x, *args, **kwargs) if x is not None else None

def apply_none(f, *args, **kwargs):
  return f(*args, **kwargs) if f is not None else None

def pprint_struct(s, indent=2, width=160):
  pp = pprint.PrettyPrinter(indent=indent, width=width)
  pp.pprint(s._to_dicts())


def pformat_struct(s, indent=2, width=160):
  pp = pprint.PrettyPrinter(indent=indent, width=width)
  return pp.pformat(s._to_dicts())


def sum_list(xs):
  assert len(xs) > 0
  return reduce(operator.add, xs)


def append_dict(d, k, v):
  xs = d.get(k) or []
  xs.append(v)

  d[k] = xs
  return d


def transpose_dicts(d):
  r = {}
  for k, v in d.items():
    for j, u in v.items():
      inner = r.get(j) or {}
      inner[k] = u
      r[j] = inner
  return r


def transpose_dict_lists(d):
  n = max([len(v) for v in d.values()])
  r = [{} for _ in range(n)]

  for k, v in d.items():
    for j, u in enumerate(v):
      r[j][k] = u
  return r


def transpose_list_dicts(xs):
  r = {}
  for d in xs:
    for k, v in d.items():
      inner = r.get(k) or []
      inner.append(v)
      r[k] = inner
  return r


def choose(*options):
  for x in options:
    if x is not None:
      return x
  assert False, "choose: all options were None"

def when(x, value):
  return None if x is None else value

def add_dict(d, k):
  d[k] = d[k] + 1 if k in d else 1
  return d


def count_dict(xs):
  counts = {}
  for k in xs:
    add_dict(counts, k)

  return counts


def sum_dicts(ds):
  r = {}

  for d in ds:
    for k, v in d.items():
      r[k] = r.get(k, 0) + v

  return r


def partition_by(xs, f):
  partitions = {}

  for x in xs:
    k, v = f(x)
    append_dict(partitions, k, v)

  return partitions


def map_type(data, data_type, f, *args, **kwargs):
  def rec(x):

    if isinstance(x, data_type):
      return f(x, *args, **kwargs)
    elif isinstance(x, str):
      return x
    elif hasattr(x, '_map'):
      return x._map(rec)
    elif isinstance(x, Mapping):
      return x.__class__({k: rec(v) for k, v in x.items()})
    elif isinstance(x, Sequence):
      return x.__class__(map(rec, x))      
    else:
      return x
  return rec(data)


def traverse_type(data, data_type, f, *args, **kwargs):
  def rec(x):

    if isinstance(x, data_type):
      f(x, *args, **kwargs)
    elif isinstance(x, str):
      pass
    elif isinstance(x, Sequence):
      for v in x:
        rec(v)
    elif isinstance(x, Mapping):
      for v in x.values():
        rec(v)

  rec(data)


def reduce_type(data, data_type, f, op, initializer=None):
  def rec(x):

    if isinstance(x, data_type):
      return f(x)
    elif isinstance(x, Sequence):
      x = [rec(v) for v in x]
      return reduce(op, x, initializer)
    elif isinstance(x, Mapping):
      x = [rec(v) for v in x.values()]
      return reduce(op, x, initializer)

  return rec(data)



from typing import List
from collections import Counter
from collections.abc import Mapping, Sequence

import torch
import numpy as np

from numbers import Number
import math

import operator
import itertools

from .struct import struct, Struct, transpose_structs

from functools import partial

class Table(Struct):
    def __init__(self, d:dict):

        assert len(d) > 0, "empty Table"
        t = next(iter(d.values()))
      
        for (k, v) in d.items():
            assert type(v) == torch.Tensor, "expected tensor, got " + type(t).__name__
            assert v.size(0) == t.size(0), "mismatched column sizes: " + str(shape(d))

        super(Table, self).__init__(d) 

    @staticmethod
    def from_structs(structs : List[Struct]) -> 'Table':
        struct_lists = transpose_structs(structs)
        return Table(struct_lists._map(torch.stack).__dict__)
   
    @staticmethod 
    def build(**d):
        return Table(d)

    def __getitem__(self, index:int) -> torch.Tensor:
        return self.__dict__[index]        

    def _index_select(self, index:torch.Tensor) -> 'Table':
        if type(index) is torch.Tensor:
            assert index.dtype == torch.int64 
            assert index.dim() == 1
            
            return self._map(lambda t: t[index])

        elif type(index) is int:
            return Struct({k: v[index] for k, v in self.items()})
        assert False, "Table.index_select: unsupported index type" + type(index).__name__


    def _index(self, index:int) -> torch.Tensor:
        return self._index_select(torch.tensor([index], dtype=int))

        
    def _narrow(self, start, n):
        return self._map(lambda t: t.narrow(0, start, n))


    def _take(self, n):
        return self._narrow(0, min(self._size, n))

    def _drop(self, n):
        n = min(self._size, n)
        return self._narrow(n, self._size - n)


    def _sequence(self):
        return (self._index_select(i) for i in range(0, self._size))

    def _sort_on(self, key, descending=False):
        assert key in self
        assert self[key].dim() == 1

        values, inds = self[key].sort(descending = descending)
        return Table({k: values if k == key else v[inds] for k, v in self.items()})

    @property
    def _head(self):
        return next(iter(self.__dict__.values()))

    @property
    def _size(self):
        return self._head.size(0)

    @property
    def _device(self):
        return self._head.device

    def _to(self, device):
        return self._map(lambda t: t.to(device))

    def _cpu(self):
        return self._map(lambda t: t.cpu())



def table(**d):
    return Table(d)


class Histogram:
    def __init__(self, values = torch.FloatTensor(0), range = (0, 1), num_bins = 10, trim = True):
        assert len(range) == 2

        self.range = range
        lower, upper = self.range

        bin_indexes = (values - lower) * num_bins / (upper - lower) 
        bin_indexes = bin_indexes.long()

        if trim:
            valid = (bin_indexes >= 0) & (bin_indexes < num_bins)

            values = values[valid]
            bin_indexes = bin_indexes[valid]
        
        bin_indexes.clamp_(0, num_bins - 1)

        self.sum         = values.sum().item()
        self.sum_squares = values.norm(2).item()
        self.counts = torch.bincount(bin_indexes, minlength = num_bins)

    def __repr__(self):
        return self.counts.tolist().__repr__()

    def bins(self):
        lower, upper = self.range
        d = (upper - lower) / self.counts.size(0)

        return torch.FloatTensor([lower + i * d for i in range(0, self.counts.size(0) + 1)])

    def to_struct(self):
        return struct(sum=self.sum, sum_squares=self.sum_squares, counts=self.counts)

    def __add__(self, other):
        assert isinstance(other, Histogram)
        assert other.counts.size(0) == self.counts.size(0), "mismatched histogram sizes"
        
        total = Histogram(range = self.range, num_bins = self.counts.size(0))
        total.sum = self.sum + other.sum
        total.sum_squares = self.sum_squares + other.sum_squares
        total.counts = self.counts + other.counts

        return total


    def __truediv__(self, x):
        assert isinstance(x, Number)

        total = Histogram(range = self.range, num_bins =  self.counts.size(0))
        total.sum = self.sum / x
        total.sum_squares = self.sum_squares / x
        total.counts = self.counts / x

        return total

    @property
    def std(self):

        n = self.counts.sum().item()
        if n > 1:
            sum_squares = self.sum_squares - (self.sum * self.sum / n)
            var = max(0, sum_squares / (n - 1))

            return math.sqrt(var)
        else:
            return 0

    @property
    def mean(self):
        n = self.counts.sum().item()
        if n > 0:
            return self.sum / self.counts.sum().item()
        else:
            return 0




def map_tensors(data, f, *args, **kwargs):
    return map_type(data, torch.Tensor, partial(f, *args, **kwargs))  

def shape_info(x):
    return map_tensors(x, lambda x: tuple([*x.shape, type(x), x.dtype]))

def shape(x):
    return map_tensors(x, lambda x: tuple(x.shape))


def from_numpy(data):
    return map_type(data, np.ndarray, torch.Tensor.from_numpy)  

def to_numpy(data):
    return map_type(data, torch.Tensor, torch.Tensor.numpy)  



def tensors_to(t, **kwargs):
    return map_tensors(t, Tensor.to, **kwargs)        


def stack_tables(tables, dim=0):
    t = transpose_structs(tables)
    return Table(dict(t._map(torch.stack, dim=dim))) 

def cat_tables(tables, dim=0):
    t = transpose_structs(tables)
    return Table(dict(t._map(torch.cat, dim=dim))) 


def split_table(table, splits):
    split = {k: v.split(splits) for k, v in table.items()}

    def build_table(i):
        return Table({k : v[i] for k, v in split.items()})

    return [build_table(i) for i in range(len(splits))]



from os.path import commonprefix
from typing import List, Tuple
from collections import Counter
from collections.abc import Mapping
import numpy as np

from numbers import Number, Integral
import math

import operator
import itertools
from functools import partial

from .struct import struct, Struct, transpose_structs, map_type, traverse_type, reduce_type


class Table(Struct):
    def __init__(self, d:dict):

        assert len(d) > 0, "empty Table"
        t = next(iter(d.values()))

        for (k, v) in d.items():
            assert hasattr(v, 'shape'), "expected ndarray, got " + type(t).__name__

        shapes = [t.shape for t in d.values()]
        self._prefix = commonprefix(shapes)

        super(Table, self).__init__(d) 

    @staticmethod
    def from_structs(structs : List[Struct]) -> 'Table':
        struct_lists = transpose_structs(structs)
        return Table(struct_lists._map(np.stack)._entries)

    @staticmethod
    def singleton(struct : Struct) -> 'Table':     
        return Table(struct._map(np.expand_dims, axis=0))

    @property
    def _shapes(self) -> 'Struct':     
        return self._map(lambda x: x.shape)

    @property
    def _shape(self) -> Tuple:
        return self._prefix

    @staticmethod 
    def stack(structs : List[Struct]) -> 'Table':
        return Table.from_structs(structs)

    @staticmethod
    def build(d:dict):
        is_array = [isinstance(t, np.ndarray) for t in d.values()]
        if all(is_array):
            return Table(d)
        else:
            return Struct(d)

    @staticmethod
    def create(**d):
        return Table(d)

    def _index_select(self, index:np.ndarray, axis=0) -> 'Table':
        assert axis < len(self._prefix)

        if isinstance(index, np.ndarray):
            assert issubclass(index.dtype.type, Integral)
            assert index.dim() == 1
        else:
            assert issubclass(type(index), Integral),\
                "Table.index_select: unsupported index type" + type(index).__name__

        return self._map(lambda t: np.take(t, index, axis=axis))

        
    def _narrow(self, start, n, axis=0):
        return self._index_select(np.arange(start, start + n), axis=axis)


    def _take(self, n, axis=0): 
        return self._narrow(0, min(self._prefix[axis], n), axis=axis)

    def _drop(self, n, axis=0):
        n = min(self._prefix[axis], n)
        return self._narrow(n, self._prefix[axis] - n, axis=axis)


    def _sequence(self, axis = 0):
        return (self._index_select(i, axis=axis) for i in range(0, self._prefix[axis]))

    def _sort_on(self, key, descending=False, axis=0):
        assert key in self
        assert self[key].dim() == 1

        values, inds = self[key].sort(descending = descending, axis=axis)
        return Table({k: values if k == key else np.take(v, inds, axis=axis) 
            for k, v in self.items()})

    @property
    def _head(self):
        return next(iter(self.__dict__.values()))

    @property
    def _size(self):
        return self._head.shape[0]


def table(**d):
    return Table(d)


class Histogram:
    def __init__(self, values = np.array([0]), range = (0, 1), num_bins = 10, trim = True):
        assert len(range) == 2

        self.range = range
        lower, upper = self.range

        bin_indexes = (values - lower) * num_bins / (upper - lower) 
        bin_indexes = bin_indexes.long()

        if trim:
            valid = (bin_indexes >= 0) & (bin_indexes < num_bins)

            values = values[valid]
            bin_indexes = bin_indexes[valid]
        
        bin_indexes.clamp_(0, num_bins - 1)

        self.sum         = values.sum().item()
        self.sum_squares = values.norm(2).item()
        self.counts = np.bincount(bin_indexes, minlength = num_bins)

    def __repr__(self):
        return self.counts.tolist().__repr__()

    def bins(self):
        lower, upper = self.range
        d = (upper - lower) / self.counts.shape[0]

        return np.array([lower + i * d for i in range(0, self.counts.shape[0] + 1)])

    def to_struct(self):
        return struct(sum=self.sum, sum_squares=self.sum_squares, counts=self.counts)

    def __add__(self, other):
        assert isinstance(other, Histogram)
        assert other.counts.shape[0] == self.counts.shape[0], "mismatched histogram sizes"
        
        total = Histogram(range = self.range, num_bins = self.counts.shape[0])
        total.sum = self.sum + other.sum
        total.sum_squares = self.sum_squares + other.sum_squares
        total.counts = self.counts + other.counts

        return total


    def __truediv__(self, x):
        assert isinstance(x, Number)

        total = Histogram(range = self.range, num_bins =  self.counts.shape[0])
        total.sum = self.sum / x
        total.sum_squares = self.sum_squares / x
        total.counts = self.counts / x

        return total

    @property
    def std(self):

        n = self.counts.sum().item()
        if n > 1:
            sum_squares = self.sum_squares - (self.sum * self.sum / n)
            var = max(0, sum_squares / (n - 1))

            return math.sqrt(var)
        else:
            return 0

    @property
    def mean(self):
        n = self.counts.sum().item()
        if n > 0:
            return self.sum / self.counts.sum().item()
        else:
            return 0

def map_arrays(data, f, *args, **kwargs):
    return map_type(data,  np.ndarray, partial(f, *args, **kwargs))  

def traverse_arrays(data, f, *args, **kwargs):
    return traverse_type(data,  np.ndarray, partial(f, *args, **kwargs))  

def reduce_arrays(data, f, op, initializer=None):
    return reduce_type(data,  np.ndarray, f, op, initializer=initializer)  



def shape_info(x):
    return map_arrays(x, lambda x: tuple([*x.shape, type(x), x.dtype]))

def shape(x):
    return map_arrays(x, lambda x: tuple(x.shape))

def arrays_astype(t, **kwargs):
    return map_arrays(t, np.ndarray.astype, **kwargs)        


def stack_tables(tables, dim=0):
    t = transpose_structs(tables)
    return Table(dict(t._map(np.stack, axis=dim))) 

def cat_tables(tables, dim=0):
    t = transpose_structs(tables)
    return Table(dict(t._map(np.concatenate, axis=dim))) 





from setuptools import setup, find_namespace_packages
setup(

    name="py-structs",
    version="0.2.3",
    author="Oliver Batchelor",
    author_email="saulzar@gmail.com",
    description="Python structs and tables using dot notation",
    url="https://github.com/saulzar/structs",
    packages=find_namespace_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
